==> Botan-1.3.6.tgz: Botan-1.3.6/src/nr.cpp <==
/*************************************************
* Nyberg-Rueppel Source File                     *
* (C) 1999-2003 The Botan Project                *
*************************************************/

#include <botan/nr.h>
#include <botan/numthry.h>
#include <botan/keypair.h>

namespace Botan {

/*************************************************
* NR_PublicKey Constructor                       *
*************************************************/
NR_PublicKey::NR_PublicKey(const DL_Group& domain, const BigInt& key)
   {
   dl_initialize(domain, key);
   powermod_g_p = FixedBase_Exp(group_g(), group_p());
   powermod_y_p = FixedBase_Exp(y, group_p());
   }

/*************************************************
* Nyberg-Rueppel Verification Function           *
*************************************************/
SecureVector<byte> NR_PublicKey::verify(const byte sig[], u32bit sig_len) const
   {
   const BigInt& q = group_q();
   const BigInt& p = group_p();
   if(sig_len != 2*q.bytes())
      return false;

   BigInt c(sig, q.bytes());
   BigInt d(sig + q.bytes(), q.bytes());

   if(c.is_zero() || c >= q || d >= q)
      throw Invalid_Argument(algo_name() + ": Invalid signature");

   BigInt i = (powermod_g_p(d) * powermod_y_p(c)) % p;
   BigInt f = (c - i) % q;

   return BigInt::encode(f);
   }

/*************************************************
* Decode X.509 NR parameters                     *
*************************************************/
void NR_PublicKey::BER_decode_params(DataSource& source)
   {
   group.BER_decode(source, DL_Group::ANSI_X9_57);
   }

/*************************************************
* Return the parameters in X.509 encoding        *
*************************************************/
SecureVector<byte> NR_PublicKey::DER_encode_params() const
   {
   return group.DER_encode(DL_Group::ANSI_X9_57);
   }

/*************************************************
* Decode a X.509 public key encoding             *
*************************************************/
void NR_PublicKey::BER_decode_pub(DataSource& source)
   {
   DL_Scheme_PublicKey::BER_decode_pub(source);
   powermod_g_p = FixedBase_Exp(group_g(), group_p());
   powermod_y_p = FixedBase_Exp(y, group_p());
   }

/*************************************************
* Return the name of this algorithm              *
*************************************************/
std::string NR_PublicKey::algo_name() const
   {
   return "NR";
   }

/*************************************************
* Return the maximum input size in bits          *
*************************************************/
u32bit NR_PublicKey::max_input_bits() const
   {
   return (group_q().bits() - 1);
   }

/*************************************************
* Return the size of each portion of the sig     *
*************************************************/
u32bit NR_PublicKey::message_part_size() const
   {
   const BigInt& q = group_q();
   return q.bytes();
   }

/*************************************************
* NR_PrivateKey Constructor                      *
*************************************************/
NR_PrivateKey::NR_PrivateKey(const DL_Group& domain) : NR_PublicKey(domain, 0)
   {
   x = random_integer(2, group_q() - 1, LongTermKey);
   y = powermod_g_p(x);
   powermod_y_p = FixedBase_Exp(y, group_p());

   if(!check_params())
      throw Self_Test_Failure(algo_name() + " private key generation failed");
   }

/*************************************************
* NR_PrivateKey Constructor                      *
*************************************************/
NR_PrivateKey::NR_PrivateKey(const DL_Group& domain,
                             const BigInt& priv, const BigInt& pub) :
   NR_PublicKey(domain, pub)
   {
   x = priv;
   if(x <= 1 || x >= group_q())
      throw Invalid_Argument(algo_name() + ": Invalid private key");
   }

/*************************************************
* Nyberg-Rueppel Signature Operation             *
*************************************************/
SecureVector<byte> NR_PrivateKey::sign(const byte in[], u32bit len) const
   {
   const BigInt& q = group_q();
   BigInt f;
   f.binary_decode(in, len);

   if(f >= q || f.is_negative())
      throw Invalid_Argument("NR_PrivateKey::sign: Input is too large");

   while(true)
      {
      BigInt k;
      do
         k.randomize(q.bits(), SessionKey);
      while(k >= q);

      BigInt c = (powermod_g_p(k) + f) % q;
      if(c.is_zero())
         continue;

      BigInt d = (k - x * c) % q;

      SecureVector<byte> output(2*q.bytes());
      c.binary_encode(output + (output.size() / 2 - c.bytes()));
      d.binary_encode(output + (output.size() - d.bytes()));
      return output;
      }
   }

/*************************************************
* Decode a PKCS #8 NR private key                *
*************************************************/
void NR_PrivateKey::BER_decode_priv(DataSource& source)
   {
   DL_Scheme_PrivateKey::BER_decode_priv(source);
   powermod_g_p = FixedBase_Exp(group_g(), group_p());
   y = powermod_g_p(x);
   powermod_y_p = FixedBase_Exp(y, group_p());
   }

/*************************************************
* Check Private Nyberg-Rueppel Parameters        *
*************************************************/
bool NR_PrivateKey::check_params() const
   {
   if(!DL_Scheme_PublicKey::check_params())
      return false;
   if(y != powermod_g_p(x))
      return false;

   try {
      KeyPair::check_key(get_pk_signer(*this, "EMSA1(SHA-1)"),
                         get_pk_verifier(*this, "EMSA1(SHA-1)")
         );
      }
   catch(Self_Test_Failure)
      {
      return false;
      }

   return true;
   }

}

==> Botan-1.3.14.tgz: Botan-1.3.14/src/nr.cpp <==
/*************************************************
* Nyberg-Rueppel Source File                     *
* (C) 1999-2004 The Botan Project                *
*************************************************/

#include <botan/nr.h>
#include <botan/numthry.h>
#include <botan/keypair.h>

namespace Botan {

/*************************************************
* NR_PublicKey Constructor                       *
*************************************************/
NR_PublicKey::NR_PublicKey(const DL_Group& grp, const BigInt& y1)
   {
   group = grp;
   y = y1;
   X509_load_hook();
   }

/*************************************************
* Algorithm Specific X.509 Initialization Code   *
*************************************************/
void NR_PublicKey::X509_load_hook()
   {
   core = NR_Core(group, y);
   check_loaded_public();
   }

/*************************************************
* Nyberg-Rueppel Verification Function           *
*************************************************/
SecureVector<byte> NR_PublicKey::verify(const byte sig[], u32bit sig_len) const
   {
   return core.verify(sig, sig_len);
   }

/*************************************************
* Return the maximum input size in bits          *
*************************************************/
u32bit NR_PublicKey::max_input_bits() const
   {
   return (group_q().bits() - 1);
   }

/*************************************************
* Return the size of each portion of the sig     *
*************************************************/
u32bit NR_PublicKey::message_part_size() const
   {
   return group_q().bytes();
   }

/*************************************************
* Create a NR private key                        *
*************************************************/
NR_PrivateKey::NR_PrivateKey(const DL_Group& grp)
   {
   group = grp;
   x = random_integer(2, group_q() - 1, LongTermKey);

   PKCS8_load_hook();
   check_generated_private();
   }

/*************************************************
* NR_PrivateKey Constructor                      *
*************************************************/
NR_PrivateKey::NR_PrivateKey(const DL_Group& grp, const BigInt& x1,
                             const BigInt& y1)
   {
   group = grp;
   y = y1;
   x = x1;

   PKCS8_load_hook();
   check_loaded_private();
   }

/*************************************************
* Algorithm Specific PKCS #8 Initialization Code *
*************************************************/
void NR_PrivateKey::PKCS8_load_hook()
   {
   if(y == 0)
      y = power_mod(group_g(), x, group_p());
   core = NR_Core(group, y, x);
   }

/*************************************************
* Nyberg-Rueppel Signature Operation             *
*************************************************/
SecureVector<byte> NR_PrivateKey::sign(const byte in[], u32bit length) const
   {
   const BigInt& q = group_q();

   BigInt k;
   do
      k.randomize(q.bits(), SessionKey);
   while(k >= q);

   return core.sign(in, length, k);
   }

/*************************************************
* Check Private Nyberg-Rueppel Parameters        *
*************************************************/
bool NR_PrivateKey::check_key(bool strong) const
   {
   if(!DL_Scheme_PrivateKey::check_key(strong) || x >= group_q())
      return false;

   if(!strong)
      return true;

   try {
      KeyPair::check_key(get_pk_signer(*this, "EMSA1(SHA-1)"),
                         get_pk_verifier(*this, "EMSA1(SHA-1)")
         );
      }
   catch(Self_Test_Failure)
      {
      return false;
      }

   return true;
   }

}
